function [out0, out] = allsims(p, n, N)

% p = number of variables 
% n = sample size
% N = number of simulations

% To duplicate the simulations in the paper
% in the paper: [out0, out] = allsims(30, 100, 100)

figure(1)
[out01, out1] = sims(p, n, N, 0, 0, .5);
figure(2)
[out02, out2] = sims(p, n, N, .3, .4, 0);
figure(3)
[out03, out3] = sims(p, n, N, .3, .4, .5);
figure(4)
[out04, out4] = sims(p, n, N, .3, .4, 1);
figure(5)
[out05, out5] = sims(p, n, N, .5, .2, 0);
figure(6)
[out06, out6] = sims(p, n, N, .5, .2, .5);
figure(7)
[out07, out7] = sims(p, n, N, .5, .2, 1);

disp('===========================');
out = [out1, out2, out3, out4, out5, out6, out7];
out0 = [out01, out02, out03, out04, out05, out06, out07];


%% ============================= subfunction "sims" =============================

function [out0, out] = sims(p, n, N, eps1, eps2, fracsmall)

% p = number of variables 
% n = sample size
% N = number of simulations
% eps1 = proportion of matrices with some contamination
% eps2 = proportion of possibly contaminated vectors within contaminated matrices
% fracsmall = fraction of contaminated rvs with small (=.3) sd's (the remaining
%    fraction of contamina[meanTMIL, meanMILD]ted rvs have big (= 3) sd's)

% unchangedLimit = number of consecutive generations with the same best
    % design, for termination
% crossoverProb = prob that two chosen parents will reproduce; otherwise
    % the child is a copy of the best of the parents
% init/final mutationProb = inital and final prob that the child will undergo a random mutation, in
    % each of h randomly chosen spots; varies linearly from init to final with "unchanged"
% eliteProp = proportion of population (the best members) to always survive

% Interesting output which is ready to be called, but at the moment may not be:
% meanTMIL0, meanMILD0, meanTMIL, meanMILD,  ngenTMIL, ngenMILD,
% diags_tmil, diags_mild, meanDELTA0, meanDELTA, meanOMEGA0, meanOMEGA

% Example of call: [meanTMIL0, meanMILD0, meanTMIL, meanMILD] = sims(30, 100, 100, .6, .2, .5);
% Figure 1:  [meanTMIL0, meanMILD0, meanTMIL, meanMILD] = sims(30, 100, 100, 0, 0, .5);
% Figure 2:  [meanTMIL0, meanMILD0, meanTMIL, meanMILD] = sims(30, 100, 100, .6, .2, 0);
% Figure 3:  [meanTMIL0, meanMILD0, meanTMIL, meanMILD] = sims(30, 100, 100, .6, .2, 1);
% Figure 4:  [meanTMIL0, meanMILD0, meanTMIL, meanMILD] = sims(30, 100, 100, 1, .2, .5);

popSize = n; 
unchangedLimit = 5;  % usual value = 5; set = 0 if no generations are to be generated, after the first
crossoverProb = .95;
initMutationProb = 0;
finalMutationProb = .30;
eliteProp = 2/popSize;

dodiagnostics = 0;
L = 2*p;
h = floor((n+1)/2);
fracbig = 1-fracsmall; % fraction of contamination with big variances

tic;

nclean_matrices = 2*floor(n*(1-eps1)); % number of totally clean matrices Uj, Vj (S = U+iV)
ndirty_matrices = 2*n - nclean_matrices; % number of matrices with possible contamination
ncontaminated_rvs = p*L*ndirty_matrices; % number of rvs in these matrices

% distribution of types of rvs among the contaminated matrices:
nsmall_rvs = floor(ncontaminated_rvs*fracsmall*eps2); 
nbig_rvs = floor(ncontaminated_rvs*fracbig*eps2); 
nclean_rvs = ncontaminated_rvs - nsmall_rvs - nbig_rvs;

TMIL0 = zeros(p,p,N);
MILD0 = zeros(p,p,N);
TMIL = zeros(p,p,N);
MILD = zeros(p,p,N);
ngenTMIL = zeros(1, N);
ngenMILD = zeros(1, N);
DELTA0 = zeros(N,1);
DELTA = zeros(N,1);
OMEGA0 = zeros(N,1);
OMEGA = zeros(N,1);
for count = 1:N
    % Generate a sample of n psd matrices S1,...,Sn.
    % totally clean matrices:
    cleannormals = randn(p,L,nclean_matrices/2)+ 1i*randn(p,L,nclean_matrices/2);
    % possibly contaminated normals:
    z = [randn(nclean_rvs, 1); .3*randn(nsmall_rvs, 1); 3*randn(nbig_rvs, 1)];
    [~,I] = sort(randperm(length(z)));
    z = z(I);
    Uvec = z(1:(length(z)/2));
    Vvec = z((length(z)/2 + 1):length(z));
    U = reshape(Uvec, p, L, ndirty_matrices/2);
    V = reshape(Vvec, p, L, ndirty_matrices/2);
    contaminatednormals = U + 1i*V;
    % combine these to ge the X-sample:
    X = cat(3, cleannormals, contaminatednormals);
    
    S = zeros(p,p,n);
    for j = 1:n     
        S(:,:,j) = X(:,:,j)*X(:,:,j)'/(2*L);  % Sigma0 = identity
    end
    % Compute each of the estimates
    if count == N
        dodiagnostics = 1;
    end
    if(count == 1 || rem(count,10) == 0) 
        disp(['count = ', num2str(count), '; TMIL']);
    end
    [sigmahat_tmil0, sigmahat_tmil, diags_tmil, ngen, mindelta0, mindelta] = tmil(S, n, p, h, dodiagnostics, ...
    unchangedLimit, popSize, crossoverProb, initMutationProb, finalMutationProb, eliteProp);
    TMIL0(:,:,count) = sigmahat_tmil0;
    TMIL(:,:,count) = sigmahat_tmil;
    ngenTMIL(count) = ngen;
    DELTA0(count) = mindelta0;
    DELTA(count) = mindelta;
    
    if(count == 1 || rem(count,10) == 0)
        disp(['count = ', num2str(count), '; MILD']);
    end
    [sigmahat_mild0, sigmahat_mild, diags_mild, ngen, minomega0, minomega] = mild(S, n, p, h, dodiagnostics, ...
    unchangedLimit, popSize, crossoverProb, initMutationProb, finalMutationProb, eliteProp);
    MILD0(:,:,count) = sigmahat_mild0;
    MILD(:,:,count) = sigmahat_mild;
    ngenMILD(count) = ngen;
    OMEGA0(count) = minomega0;
    OMEGA(count) = minomega;
end

% Generation one output:
meanDELTA0 = mean(DELTA0);
meanOMEGA0 = mean(OMEGA0);
meanTMIL0 = mean(TMIL0,3);
meanMILD0 = mean(MILD0,3);

msevecTMIL0 = zeros(N,1);
msevecMILD0 = zeros(N,1);
for j = 1:N
    msevecTMIL0(j) = norm(TMIL0(:,:,j) - eye(p), 'fro')^2;    
    msevecMILD0(j) = norm(MILD0(:,:,j) - eye(p), 'fro')^2;   
end

rmseTMIL0 = sqrt(mean(msevecTMIL0))/p; % per element rmse
rmseMILD0 = sqrt(mean(msevecMILD0))/p; % per element rmse 
pooledSD0 = sqrt((var(msevecTMIL0)/mean(msevecTMIL0)...
    + var(msevecMILD0)/mean(msevecMILD0))/(8*p*p*N)); % standard error 
biasTMIL0 = norm(meanTMIL0 - eye(p),'fro')/p;
biasMILD0 = norm(meanMILD0 - eye(p),'fro')/p;
out0 = [rmseTMIL0; rmseMILD0; pooledSD0; biasTMIL0; biasMILD0; meanDELTA0; meanOMEGA0];

% Final output:
meanDELTA = mean(DELTA);
meanOMEGA = mean(OMEGA);
meanTMIL = mean(TMIL,3);
meanMILD = mean(MILD,3);

msevecTMIL = zeros(N,1);
msevecMILD = zeros(N,1);
for j = 1:N
    msevecTMIL(j) = norm(TMIL(:,:,j) - eye(p), 'fro')^2;    
    msevecMILD(j) = norm(MILD(:,:,j) - eye(p), 'fro')^2;   
end

rmseTMIL = sqrt(mean(msevecTMIL))/p; % per element rmse
rmseMILD = sqrt(mean(msevecMILD))/p; % per element rmse 
pooledSD = sqrt((var(msevecTMIL)/mean(msevecTMIL)...
    + var(msevecMILD)/mean(msevecMILD))/(8*p*p*N)); % standard error 
biasTMIL = norm(meanTMIL - eye(p),'fro')/p;
biasMILD = norm(meanMILD - eye(p),'fro')/p;
out = [rmseTMIL; rmseMILD; pooledSD; biasTMIL; biasMILD; meanDELTA; meanOMEGA];

out0 = [1000*out0(1:5,:); out0(6:7,:)];
out0 = round(100*out0)/100;

out = [1000*out(1:5,:); out(6:7,:)];
out = round(100*out)/100;

toc
disp('  ');
disp(['eps1, eps2, fracsmall = ', num2str([eps1, eps2, fracsmall])]);
disp('  ');
disp('1000*rmseTMIL, 1000*rmseMILD, 1000*pooledSD, 1000*biasTMIL, 1000*biasMILD,'); 
disp('meanDELTA, meanOMEGA before and after genetic algorithm, are ');
disp('   ');
disp([out0, out]);

disp('===========================');


%% ============================= subfunction "tmil" =============================
% Computes the TMIL estimate once, from a sample of size n
function [sigmahat_tmil0, sigmahat_tmil, diags_tmil, ngen, mindelta0, mindelta] = tmil(S, n, p, h, dodiagnostics, ...
    unchangedLimit, popSize, crossoverProb, initMutationProb, finalMutationProb, eliteProp)

% Initialize:
lossvec = zeros(popSize,1); %loss of each member of the population in one generation
population = zeros(popSize, n);  % Indicators for entire current population
genmat = []; % Keep track of the BEST (of the current population) set of 
        % subsampleIndicatrors in each generation
generation = 1; 
unchanged = 0;

% Simulate an initial population of subsamples; compute the loss for each member
perm = randperm(n);
for a = 1:popSize
    % Start with closest h matrices to S(:,:,a); compute Sigmahat(1):
    [bestsubsampleIndicators, ~, ~]  = besth_delta(S, S(:,:,perm(a)), h);
	% Compute loss
    [lossvec(a),bestsubsampleIndicators] = lossdelta(bestsubsampleIndicators, S);
    population(a,:) = bestsubsampleIndicators; % a vector of n-h 0's and h 1's 
end

bestsubsampleIndicators = population(lossvec == min(lossvec), :);
if size(bestsubsampleIndicators,1) > 1
	bestsubsampleIndicators = bestsubsampleIndicators(1, :);
end
genmat = [genmat; [bestsubsampleIndicators, min(lossvec)]];
indi = 1:n;
bestsubsample = indi(bestsubsampleIndicators == 1);
sigmahat_tmil0 = mean(S(:,:,bestsubsample),3);
mindelta0 = min(lossvec);
% disp('first TMIL population and loss:');
% disp([population lossvec]);

%% Create subsequent generations
while unchanged < unchangedLimit
	generation = generation + 1;
	fitvec = fitness(lossvec); % fitness of all designs in generation
    newPopulation = zeros(popSize, n);
    % Force elites (the best of the current population) to survive
    numberElites = floor(popSize*eliteProp);
    [~,I] = sort(fitvec,'descend');
    newPopulation(1:numberElites, :) = population(I(1:numberElites),:);
    lossvec(1:numberElites) = lossvec(I(1:numberElites));
    newPopSize = numberElites;
    
    % Create the remaining children:
    cumfit = cumsum(fitvec);    
    while(newPopSize < popSize)
        % Create a child from 2 randomly chosen members of the current population        
        mutationProb = initMutationProb + ...
            (finalMutationProb-initMutationProb)*unchanged/unchangedLimit;
        child  = combine(population, fitvec, cumfit, h, n, crossoverProb, mutationProb);
        % Compute the loss and update the new population
        newPopSize = newPopSize + 1;
        [lossvec(newPopSize),bestsubsampleIndicators] = lossdelta(child, S);        
        newPopulation(newPopSize,:) = bestsubsampleIndicators;
    end % end of inner "while"     
    population = newPopulation;  % Replace the old population with the new one
    % disp('new TMIL population and loss:');
    % disp([population lossvec]);
    
    % Summarize this generation:
    bestsubsampleIndicators = population(lossvec == min(lossvec), :);
    if size(bestsubsampleIndicators,1) > 1
        bestsubsampleIndicators = bestsubsampleIndicators(1, :);
    end
    genmat = [genmat; [bestsubsampleIndicators, min(lossvec)]];
    if isequal(genmat(generation, 1:n), genmat(generation - 1, 1:n))
        unchanged = unchanged + 1;
    else
        unchanged = 0;
    end
    % [generation, unchanged]
end % end of outer "while"


%% Summarize the output
ngen = size(genmat, 1);
indi = 1:n;
bestsubsample = indi(genmat(ngen, 1:n) == 1);
sigmahat_tmil = mean(S(:,:,bestsubsample),3);
mindelta = genmat(ngen, n+1);

% disp(['# of generations = ', num2str(ngen),'  Min loss per generation = '] );
% disp(vpa(genmat(:, n+1),8));  
% disp('best subsamples in generations:');
% disp(genmat(:,1:n));

diags_tmil = [];
if (dodiagnostics == 1)
    diags = diagnostics(S, sigmahat_tmil)';
    diags_tmil = diags(1,:);
%    subplot(2,2,1);
%    plot(genmat(:, n+1));
%    title('minimum loss in generation');
    subplot(1,2,1);
    qqplot(diags_tmil(1,:));
    xlabel('(a)');
    ylabel('Sample Quantiles')
    title(' ');
%   subplot(1,3,3);
%   qqplot(diags_tmil(2,:));
%   xlabel('(c)');
%   title('qq plot; Omega with TMIL');
end

%% ============================= subfunction "mild" =============================
% computes the MILD estimate once, from a sample of size n
function [sigmahat_mild0, sigmahat_mild, diags_mild, ngen, minomega0, minomega] = mild(S, n, p, h, dodiagnostics, ...
    unchangedLimit, popSize, crossoverProb, initMutationProb, finalMutationProb, eliteProp)

% Initialize:
lossvec = zeros(popSize,1); %loss of each member of the population in one generation
population = zeros(popSize, n);  % Indicators for entire current population
genmat = []; % Keep track of the BEST (of the current population) set of 
        % subsampleIndicatrors in each generation
generation = 1; 
unchanged = 0;

% Simulate an initial population of subsamples; compute the loss for each member
perm = randperm(n);
for a = 1:popSize
    % Start with closest h matrices to S(:,:,a); compute Sigmahat(1):
    Sa = S(:,:,perm(a));
    omegavec = zeros(n,1);
    indices = 1:n;
    for j = indices
    omegavec(j) = abs(real(omega(S(:,:,j), Sa)) - p);
    end
    [~, I] = sort(omegavec);
    orderedindices = indices(I)';
    bestset = orderedindices(1:h); % indices of closest subsample
    bestsubsampleIndicators = zeros(1,n);
    bestsubsampleIndicators(bestset) = 1;  % indicators of closest subsample
	% Compute loss and resulting subsample, STARTING with this subsample
    [lossvec(a),bestsubsampleIndicators] = lossomega(bestsubsampleIndicators, S);
    population(a,:) = bestsubsampleIndicators; % a vector of n-h 0's and h 1's 
end

bestsubsampleIndicators = population(lossvec == min(lossvec), :);
if size(bestsubsampleIndicators,1) > 1
	bestsubsampleIndicators = bestsubsampleIndicators(1, :);
end
genmat = [genmat; [bestsubsampleIndicators, min(lossvec)]];
indi = 1:n;
bestsubsample = indi(bestsubsampleIndicators == 1);
sigmahat_mild0 = mean(S(:,:,bestsubsample),3);
minomega0 = min(lossvec)^(1/p);  % Determinant^(1/p)
% disp('first MILD population and loss:');
% disp([population lossvec]);

%% Create subsequent generations
while unchanged < unchangedLimit
	generation = generation + 1;
	fitvec = fitness(lossvec); % fitness of all designs in generation
    newPopulation = zeros(popSize, n);
    % Force elites (the best of the current population) to survive
    numberElites = floor(popSize*eliteProp);
    [~,I] = sort(fitvec,'descend');
    newPopulation(1:numberElites, :) = population(I(1:numberElites),:);
    lossvec(1:numberElites) = lossvec(I(1:numberElites));
    newPopSize = numberElites;
    
    % Create the remaining children:
    cumfit = cumsum(fitvec);    
    while(newPopSize < popSize)
        % Create a child from 2 randomly chosen members of the current population        
        mutationProb = initMutationProb + ...
            (finalMutationProb-initMutationProb)*unchanged/unchangedLimit;
        child  = combine(population, fitvec, cumfit, h, n, crossoverProb, mutationProb);
        % Compute the loss and update the new population
        newPopSize = newPopSize + 1;
        [lossvec(newPopSize),bestsubsampleIndicators] = lossomega(child, S);        
        newPopulation(newPopSize,:) = bestsubsampleIndicators;
    end % end of inner "while"     
    population = newPopulation;  % Replace the old population with the new one
    % disp('new MILD population and loss:');
    % disp([population lossvec]);
    
    % Summarize this generation:
    bestsubsampleIndicators = population(lossvec == min(lossvec), :);
    if size(bestsubsampleIndicators,1) > 1
        bestsubsampleIndicators = bestsubsampleIndicators(1, :);
    end
    genmat = [genmat; [bestsubsampleIndicators, min(lossvec)]];
    if isequal(genmat(generation, 1:n), genmat(generation - 1, 1:n))
        unchanged = unchanged + 1;
    else
        unchanged = 0;
    end
    % [generation, unchanged]
end % end of outer "while"


%% Summarize the output
ngen = size(genmat, 1);
indi = 1:n;
bestsubsample = indi(genmat(ngen, 1:n) == 1);
sigmahat_mild = mean(S(:,:,bestsubsample),3);
minomega = genmat(ngen, n+1)^(1/p);   %  Determinant^(1/p)

% disp(['# of generations = ', num2str(ngen),'  Min loss per generation = '] );
% disp(vpa(genmat(:, n+1),8));  
% disp('best subsamples in generations:');
% disp(genmat(:,1:n));

diags_mild = [];
if (dodiagnostics == 1)
    diags = diagnostics(S, sigmahat_mild)';
    diags_mild = diags(2,:);
%    subplot(2,2,3);
%    plot(genmat(:, n+1));
%    title('minimum loss in generation');
    subplot(1,2,2);
    qqplot(diags_mild);
    xlabel('(b)');    
    ylabel('Sample Quantiles')
    title(' ');
end


%% ===============================================================================
%% ===============================================================================

%% ============================= subfunction "lossdelta" =============================
function [loss, bestsubsampleIndicators] = lossdelta(subsampleIndicators, S)
% Carries out the iterative steps of algorithm 1, given a starting set of
% subsampleIndicators
h = sum(subsampleIndicators);
n = length(subsampleIndicators);
indi = 1:n;
subsample = indi(subsampleIndicators == 1);
Sample = S(:,:,subsample);
Sbar = mean(Sample,3);
delbar = deltabar(Sample, Sbar);
change = 1;
while change > 0,
	olddel = delbar;
	[bestIndicators, bestS, Sbar]  = besth_delta(S, Sbar, h);
	delbar = deltabar(bestS, Sbar);
	change = olddel - delbar;
end
loss = delbar;
bestsubsampleIndicators = bestIndicators;
%% ============================= subfunction "deltabar" =============================
% Computes deltabar, given a sample 'Sample' and its mean 'Sbar'
function delbar = deltabar(Sample, Sbar)
h = size(Sample,3);
Sdet = zeros(h,1);
for j = 1:h
	Sdet(j) = det(Sample(:,:,j));
end
delbar = real(log(det(Sbar)) - mean(log(Sdet)));

%% ============================= subfunction "besth_delta" =============================
% Returns the h matrices closest to Sbar, and their average
function [bestIndicators, bestS, Sbar]  = besth_delta(S,Sbar,h)
n = size(S,3);
delvec = zeros(n,1);
indices = 1:n;
for j = indices
    delvec(j) = real(delta(S(:,:,j), Sbar));
end
[~, I] = sort(delvec);
orderedindices = indices(I)';
bestset = orderedindices(1:h); % indices of closest subsample
bestS = S(:,:,bestset);  % closest subsample to input Sbar
bestIndicators = zeros(1,n);
bestIndicators(bestset) = 1;  % indicators of closest subsample
Sbar = mean(bestS,3);  % output Sbar

%% ============================= subfunction "delta" =============================
function delta  = delta(Sj,Sbar) 
M = Sbar\Sj;
delta = trace(M) - log(det(M)) - size(M,1);

%% ============================= subfunction "lossomega" =============================
function [loss, bestsubsampleIndicators] = lossomega(subsampleIndicators, S)
% Carries out the iterative steps of algorithm 1, given a starting set of
% subsampleIndicators
h = sum(subsampleIndicators);
n = length(subsampleIndicators);
indi = 1:n;
subsample = indi(subsampleIndicators == 1);
Sample = S(:,:,subsample);
Sbar = mean(Sample,3);
omegabar = real(det(Sbar));
change = 1;
while change > 0,
	oldomega = omegabar;
	[bestIndicators, ~, Sbar]   = besth_omega(S, Sbar, h);
	omegabar = real(det(Sbar));
	change = oldomega - omegabar;    
end
loss = omegabar;
bestsubsampleIndicators = bestIndicators;
%% ============================= subfunction "besth_omega" =============================
% Returns the h matrices closest to Sbar, their average and their indices
function [bestIndicators, bestS, Sbar]  = besth_omega(S,Sbar,h)
n = size(S,3);
omegavec = zeros(n,1);
indices = 1:n;
for j = indices
    omegavec(j) = real(omega(S(:,:,j), Sbar));
end
[~, I] = sort(omegavec);
orderedindices = indices(I)';
bestset = orderedindices(1:h); % indices of closest subsample
bestS = S(:,:,bestset);  % closest subsample to input Sbar
bestIndicators = zeros(1,n);
bestIndicators(bestset) = 1;  % indicators of closest subsample
Sbar = mean(bestS,3);  % output Sbar

%% ============================= subfunction "omega" =============================
function omega  = omega(Sj,Sbar) 
M = Sbar\Sj;
omega = trace(M);

%% ============================= subfunction "fitness" =============================
function fit = fitness(x) 
fit = 1./(x - .999*min(x)).^2;
fit = fit/sum(fit);

%% ============================= subfunction "combine" =============================
function child  = combine(population, fitvec, cumfit, h, n, crossoverProb, mutationProb) 
popIndices = 1:size(population, 1);
indi = 1:n;
choices = zeros(2,1);

rr = rand(2,1);
for i = 1:2
	choices(i) = min(popIndices(cumfit >= rr(i)));
end

parents = population(choices, :);

if rand <= crossoverProb  
	% With this prob, form a child from the two subsamples chosen above
	child = parents(1,:) + parents(2,:);
	singletonsToDrop = length(child(child > 0)) - h; 
	% drop some sample members which appear only once:
        if singletonsToDrop > 0
            singletons = indi(child == 1);
            ss = randperm(length(singletons));
            singletons = singletons(ss); % permute the singletons
            indicesToDrop = singletons(1:singletonsToDrop);
            child(indicesToDrop) = 0;
        end
	child(child == 2) = 1; % keep any point which appears twice
else % otherwise the child is identical to the best of the parents
	if fitvec(choices(1)) > fitvec(choices(2))
        child = parents(1,:);
    else
        child = parents(2,:);
	end
end
numberOfMutations = 0;
while numberOfMutations <= h  % Possibly cause up to h mutations
	if rand < mutationProb
        s1 = randperm(h);
        s0 = randperm(n-h);
        child1 = indi(child==1);
        child0 = indi(child==0);
        child(child1(s1(1))) = 0;
        child(child0(s0(1))) = 1;    
	end
	numberOfMutations = numberOfMutations + 1;
end

%% ============================= subfunction "diagnostics" =============================
%First columns is {t_j} = Delta, second is {d_j} = Omega
function diags = diagnostics(S, sigmahat)
n = size(S,3);
p = size(S,2);
diags = zeros(n,2);
for j = 1:n
	diags(j,:) = [real(trace(sigmahat\S(:,:,j)) - log(det(sigmahat\S(:,:,j)))) - p, real(trace(sigmahat\S(:,:,j)))];
end
