#mild -- minimum information loss estimate. X is a p x p x n, 3-dimensional array containing the sample of n matrices, 
#	0.1 is the default proportion to trim. 
mild <- function( X, prop=0.1 )
	{
	np <- length( X[,1,1] )			# np is the matrix dimension
	nmat <- length( X[1,1,] )		# nmat is the no. of matrices in the sample
	h <- round( nmat * (1.0 - prop) )	# h is the trimmed sample size


	sx <- seq( 1, nmat, 1 )			# sx is the matrix index vector, 1,...,nmat
	mindet <- 10^6				# used to check for minimum determinant

	for( j in 1:nmat )			# Step through each matrix
		{
		k <- 0				# Counter for iterations
		dk <- mdx( X[,,j], X )		# Calculate the distance for each matrix relative to X[,,j]

		gix <- rank( dk ) <= h		# pick the h smallest deltas
		SH <-  meanmat( X[,,gix] )	# mean of the h matrices picked
		dh <- log(detz( SH ) )		# dh is the initial log(determinant) of SH

		dhold <- 0.0
		dhdiff <- abs( dh - dhold )	# inital value for difference between iterations
		print( paste( 'Matrix ', j ),quote=F)
 		print( paste(k, dh,' diff = ', abs( dh - dhold ) ), quote=F )

		while ( dhdiff > 0.0  )
 			{ 
			dk <- mdx( SH, X )	 	# calculates distances for the entire sample using mean of 
								# S from trimmed sample
			gix <- rank( dk ) <= h		# pick the smallest dk's
			dhold <- dh				# save old value dh
			SH <-  meanmat( X[,,gix] )	# update SH
			dh <- log( detz( SH ) )		# update dh

			dhdiff <- abs( dh - dhold )	# update difference between iterations
			k <- k + 1
 			print( paste(k, dh,' diff = ', abs( dh - dhold ) ), quote=F )
			}
		if( dh < mindet )				# save matrix set (mingix) where dh is minimum  
			{
			mindet <- dh    
			mingix <- gix
			}
		}

	MILD <- meanmat( X[,,mingix] )
	sxz <- sx[mingix]
	trmix <- sx[!mingix ]
							# returns a list where z$MILD is minimum MILD est
							# z$mindet is minimum log(determinant) value
							# z$sxz contains indices of the trimmed sample
							# z$trmix are the observations trimmed 
	z <- list( MILD=MILD, mindet = mindet, trmix=trmix, sxz=sxz )
	z
	}
#mdx -- fuction to calculate an array of tr( inv(SM) %*% Si ), i=1,...,n
mdx <- function( SM, S )
	{
	np <- length( S[1,,1] )
	nm <- length( S[1,1,] )
	dk <- rep(0, nm) 	# array for distance values.

	SMI <- solve( SM )
	TX <- array( apply( S, 3, FUN= "%*%", SMI ), c( np, np, nm ) )
	dk <- Re( apply( TX, 3, tr ) )   	# array of tr( inv(mean( S ) ) * S[,,i] ), i=1,...,nm
	dk
	}
# detz -- calculates the determinant of a square matrix.
detz <- function( A )
 	{ 
	qq <- qr( A )
	R <- qr.R( qq )
	abs( prod( diag( R ) ) )
	}
#meanmat -- calculates the mean of a set of square complex-valued matrices.
meanmat <- function( X )
	{
	np <- length( X[1,,1] )		# matrix dimension
	ns <- length( X[1,1,] )		# number of layers ( observations )
	
	XS <- matrix(0+0i, np, np )
	for( i in 1:ns )
			XS <- XS + X[,,i] 
	XS <- XS / ns
	XS
	}
#tr -- matrix trace
tr <- function( W )
     {
     sum(diag( W ))
     }

