#tmil -- minimum information loss estimate. X is a p x p x n, 3-dimensional array containing the sample of n matrices, 
#	0.1 is the default proportion to trim. 
tmil <- function( X, prop=0.1 )
	{
	np <- length( X[,1,1] )			# np is the matrix dimension
	nmat <- length( X[1,1,] )		# nmat is the no. of matrices in the sample
	h <- round( nmat * (1.0 - prop) )	# h is the trimmed sample size


	sx <- seq( 1, nmat, 1 )			# sx is the matrix index vector, 1,...,nmat
	mindh <- 10^6				# used to check for minimum mean delta value

	for( j in 1:nmat )			# Step through each matrix
		{
		k <- 0				# Counter for iterations
		dk <- deltas( X[,,j], X )	# Calculate the deltas for each matrix relative to X[,,j]
		gix <- rank( dk ) <= h		# pick the h smallest deltas

		SH <-  meanmat( X[,,gix] )	# mean of the h matrices picked
		dh <- deltabar( SH, X[,,gix] )# dh is an initial mean delta value as in eqn.(3)

		dhold <- 0.0
		dhdiff <- abs( dh - dhold )	# inital value for difference between iterations
		print( paste( 'Matrix ', j ),quote=F)
 		print( paste(k, dh,' diff = ', abs( dh - dhold ) ), quote=F )

		while ( dhdiff > 0.0  )
 			{ 
			dk <- Re( deltas( SH, X ) ) 	# calculates deltas for the entire sample using mean of 
								# S from trimmed sample

			gix <- rank( dk ) <= h		# pick the smallest dk's
			dhold <- dh				# save old value of mean delta
			SH <-  meanmat( X[,,gix] )	# update SH
			dh <- deltabar( SH, X[,,gix] )# update mean delta


			dhdiff <- abs( dh - dhold )	# update difference between iterations
			k <- k + 1
 			print( paste(k, dh,' diff = ', abs( dh - dhold ) ), quote=F )
			}
		if( dh < mindh )			# save matrix set (mingix) where dh is minimum  
			{
			mindh <- dh    
			mingix <- gix
			}
		}

	TMIL <- meanmat( X[,,mingix] )
	sxz <- sx[mingix]
	trmix <- sx[!mingix ]
							# returns a list where z$TMIL is minimum TMIL
							# Z$mindh is minimum delta in eqn 3
							# z$sxz contains indices of the trimmed sample
							# z$trmix are the observations trimmed 
	z <- list( TMIL=TMIL, mindh = mindh, trmix=trmix, sxz=sxz )
	z
	}

deltas <- function( SM, S )
        {
        np <- length( S[1,,1] )
        nm <- length( S[1,1,] )
        dk <- rep(0, nm)                # array for divergence values
        SMI <- solve( SM )
        for( j in 1:nm )
                {
                W <- SMI %*% S[,,j]
                dk[j] <- Re( tr( W ) ) - log( Re( detz( W ) ) ) - np
                }
        dk
        }

# deltabar -- calculates mean KL discrimination measure relative to a matrix SM,
#	  for a set of matrices contained in a 3-D array ( S ). 
deltabar <- function(SM, S)
	{
	np <- length( S[1,,1] )
	h <- length( S[1,1,] )
	Sdet <- rep(0,h)
	for( j in 1:h )
		Sdet[j] <- detz( S[,,j] )
	z <- Re( log( detz( SM ) ) ) - mean( log( Sdet ) )
	z
	}
# detz -- calculates the determinant of a square, positive semidefinite  matrix.
detz <- function( A )
 	{ 	
	qq = qr(A)
	R = qr.R(qq)
	abs(prod(diag(R)))
	# prod( eigen(A, only.values = TRUE)[[1]] ) 
	}
#meanmat -- calculates the mean of a set of square complex-valued matrices.
meanmat <- function( X )
	{
	np <- length( X[1,,1] )		# matrix dimension
	ns <- length( X[1,1,] )		# number of layers ( observations )
	
	XS <- matrix(0+0i, np, np )
	for( i in 1:ns )
			XS <- XS + X[,,i] 
	XS <- XS / ns
	XS
	}

#tr -- matrix trace
tr <- function( W )
	{
	sum(diag( W )) 
	}
